//@author: microdee
//@help: a raymarcher template
//@tags: raymarching
//@credits: Kian Bashiri, Ińigo Quílez

// shaders and rendering are based on:
// http://www.mazapan.se/news/2010/07/15/gpu-ray-marching-with-distance-fields/
// operations and modifiers are based on:
// http://www.iquilezles.org/www/articles/distfunctions/distfunctions.htm

float4x4 tV: VIEWINVERSE;

const float EPS <string uiname = "Primary Epsilon"; float uimin=0.001;> = 0.02;
const float sEPS <string uiname = "Secondary Epsilon"; float uimin=0.001;> = 0.1;
const float PI = 3.1415926535f;
const float PI2 = 3.1415926535f *2.0f;

const bool debugmode = 0;
const float debugscale = 0.01;

float FOVlength <string uiname = "FOV";> = 1;
float ScreenRatio = 1;
const float maxDist <string uiname = "Maximum Distance";> = 500;
const int maxIterations <string uiname = "Maximum Iterations";> = 150;

float3 Position = float3(0, 10, 10);
float3 DirectionX = float3(1, 0, 0);
float3 DirectionY = float3(0, 1, 0);
float3 DirectionZ = float3(0, 0, 1);

const float4 skyColor : COLOR = float4 ( 176, 215, 225, 255 ) /255.0f;

#include "distanceField.fx"
#include "shaders.fx"
#include "materials.fx"

const float4 sunColor : COLOR = float4 ( 1.0f, 0.9f, 0.7f, 1.0f );
float3 SunDirection = {0, -5, 2};
float specPow <string uiname = "Specular Power";> = 480;

const float4 vignetteColor : COLOR = float4 ( 189, 108, 181, 255 ) /255.0f;
float3 brightness = 1;
float3 contrast = .7;

// VS_INPUT
struct VS_INPUT
{
    float2 Pos	: POSITION;
	float2 Tex	: TEXCOORD;
};

// PS_INPUT
struct PS_INPUT
{
    float4 Pos	: POSITION;
	float2 Tex	: TEXCOORD;
};

//////////////////// VertexShader ////////////////////

PS_INPUT vertexShader ( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
	
    output.Pos = float4 ( input.Pos.xy*2, 0, 1 );
	output.Tex = input.Tex;
    
    return output;
}

//////////////////// PixelShader ////////////////////

struct col
{
	float4 outcol : COLOR;
	float4 position : COLOR1;
	float4 normal : COLOR2;
	float4 debug : COLOR3;
};
col pixelShader ( PS_INPUT input)
{
	// setup ray
	float3 rayDir = float3 ( 0, 0, 0 );
	rayDir += DirectionX *( input.Tex.x *+2.0f -1.0f );
	rayDir += DirectionY *( input.Tex.y *-2.0f +1.0f ) *ScreenRatio;
	rayDir += DirectionZ *FOVlength;
	
	float3 rayPosition = Position +rayDir;
	rayDir = normalize ( rayDir );

	// march
	float3 color = skyColor;
	int materialID = -1;
	float totalDistance = 0.0f;
	int totalIterations = 0;
	int totalIterationsD = 0;
	float3 OutPos = 0;
	float3 OutNorm = 0;
	while ( totalDistance < maxDist && ++totalIterations < maxIterations )
	{
		float distance = distanceField ( rayPosition, materialID );
		if ( distance < EPS *0.5f )
		{
			OutPos = rayPosition;
			float3 N = NormalAtPoint ( rayPosition, EPS );
			OutNorm = N;
			// point light would be easy i was just lazy :P
			float3 L = SunDirection;
			float3 V = normalize ( Position -rayPosition );
			float3 H = normalize ( V +L );
			
			if ( materialID == MATERIAL_FLOOR )
				color = floorColor;
			if ( materialID == MATERIAL_BOX )
				color = boxColor;
			if ( materialID == MATERIAL_SPHERE )
				color = sphereColor;
			
			// calculate lighting
			float diffuseTerm = saturate ( dot ( N, L ) );
			color = lerp ( color, sunColor, diffuseTerm );

			shaderdata shadTemp = (shaderdata)0;
			shadTemp = AmbientOcclusion( rayPosition, N);
			float aoTerm = shadTemp.lightness;
			//color = lerp ( float3 ( 0, 0, 0 ), color, aoTerm );
			color *= aoTerm;
			totalIterationsD += shadTemp.operations;
			
			float specularTerm = saturate ( pow ( saturate ( dot ( N, H ) ), specPow ) );
			color += sunColor *specularTerm *aoTerm;
			
			shadTemp = Shadows ( rayPosition, SunDirection );
			float ShadTerm = shadTemp.lightness;
			color *= ShadTerm;
			totalIterationsD += shadTemp.operations;
			
			if ( materialID == MATERIAL_SPHERE )
			{
				shadTemp = SubSurfaceScattering ( rayPosition, rayDir);
				float sssTerm = shadTemp.lightness;
				color = lerp ( color, sphereSSSColor, sssTerm );
				totalIterationsD += shadTemp.operations;
			}
			
			break;
		}
		
		// march on
		rayPosition += rayDir *distance;
		totalDistance += distance;
		totalIterationsD++;
	}
	
	// fog
	color = postFog ( color, rayPosition -Position, SunDirection );
	
	// color correction
	color = lerp ( color, sqrt ( color ), -contrast ) * brightness;
	
	// vignette
	float vignette = pow ( saturate ( length ( input.Tex -float2 ( 0.5, 0.5f ) ) *1.6f ), 2.0f ) *0.4f;
	color = lerp ( color, vignetteColor, vignette );
	
	col final = (col)0;
	final.outcol = float4(color, 1.0f);
	final.position = float4(OutPos, 1);
	final.normal = float4(OutNorm, 1);
	final.debug = (!debugmode) ? float4 (totalIterationsD.xxx*debugscale, 1) : float4 (totalDistance.xxx*debugscale, 1);
	
	return final;
}

technique Main
{
    pass P0
    {
        VertexShader = compile vs_3_0 vertexShader();
        PixelShader = compile ps_3_0 pixelShader();
    }
}